#ifndef _VIEWER_H
#define _VIEWER_H

// Copyright (c) OpenMedia Lab., 2013. ver.8/14/2013

#include "geometry.h"
#include "keystate.h"

class Viewer{
public:
	Vector eye;
	Vector target;
	Vector up;
	Vector zenith; // at right angles with viewDirection=target-eye, different from up

	Viewer(){
	}
	void window2world( double windowX, double windowY, double windowZ, double &worldX, double &worldY, double &worldZ ){
		GLdouble mm[16], pm[16];
		GLint v[4];
		glGetDoublev( GL_MODELVIEW_MATRIX,  mm );
		glGetDoublev( GL_PROJECTION_MATRIX, pm );
		glGetIntegerv( GL_VIEWPORT, v );
		gluUnProject( windowX, double(v[3])-windowY, windowZ, mm, pm, v, &worldX, &worldY, &worldZ );
	}
	void window2world( int windowX, int windowY, double windowZ, double &worldX, double &worldY, double &worldZ ){ this->window2world( double(windowX), double(windowY), windowZ, worldX, worldY, worldZ ); }
	void window2world( int windowX, int windowY, double windowZ, Vector &world ){ this->window2world( double(windowX), double(windowY), windowZ, world.x, world.y, world.z ); }
	void window2world( double windowX, double windowY, double &worldX, double &worldY ){
		double worldZ;
		this->window2world( windowX, windowY, 0.0, worldX, worldY, worldZ );
	}
	void window2world( int windowX, int windowY, double &worldX, double &worldY ){ this->window2world( double(windowX), double(windowY), worldX, worldY ); }
	void world2window( double worldX, double worldY, double worldZ, double &windowX, double &windowY, double &windowZ ){
		GLdouble mm[16], pm[16];
		GLint v[4];
		glGetDoublev( GL_MODELVIEW_MATRIX,  mm );
		glGetDoublev( GL_PROJECTION_MATRIX, pm );
		glGetIntegerv( GL_VIEWPORT, v );
		gluProject( worldX, worldY, worldZ, mm, pm, v, &windowX, &windowY, &windowZ );
		windowY = v[3]- windowY;
	}
	void world2window( double worldX, double worldY, double worldZ, int &windowX, int &windowY, double &windowZ ){ double x, y; this->world2window( worldX, worldY, worldZ, x, y, windowZ ); windowX = int(x+0.5); windowY = int(y+0.5); }
	void world2window( Vector &world, int &windowX, int &windowY, double &windowZ ){ double x, y; this->world2window( world.x, world.y, world.z, x, y, windowZ ); windowX = int(x+0.5); windowY = int(y+0.5); }
	void world2window( double worldX, double worldY, double &windowX, double &windowY ){
		double windowZ;
		this->world2window( worldX, worldY, 0.0, windowX, windowY, windowZ );
	}
	void world2window( double worldX, double worldY, int &windowX, int &windowY ){ double x, y; this->world2window( worldX, worldY, x, y ); windowX = int(x+0.5); windowY = int(y+0.5); }
	Line cursorLine( int x, int y ){
		Vector a, b;	
		this->window2world( double(x),double(y), -1.0, a.x, a.y, a.z );
		this->window2world( double(x),double(y),  1.0, b.x, b.y, b.z );
		Line line( a, b );
		return line;
	}
	Vector viewDirection(){ Vector v; v.subtract(target, eye); return v; }
	Vector rightDirection(){ Vector v; v.exterior(this->viewDirection(), up); return v; }
	void lookAt(double eyeX,double eyeY, double eyeZ, double targetX, double targetY, double targetZ, double upX, double upY, double upZ){
		eye.set(eyeX, eyeY, eyeZ);
		target.set(targetX, targetY, targetZ);
		up.set(upX, upY, upZ);
		up.normalize();
		Vector viewDirection = target - eye;
		Vector leftDirection = up & viewDirection;
		zenith = viewDirection & leftDirection;
		zenith.normalize();
		glMatrixMode( GL_MODELVIEW );
		glLoadIdentity();
		gluLookAt( eyeX, eyeY, eyeZ, targetX, targetY, targetZ, upX, upY, upZ );
	}
	void lookAt(Vector &eye, Vector &target, Vector &up){ this->lookAt( eye.x, eye.y, eye.z, target.x, target.y, target.z, up.x, up.y, up.z ); }
	Line ViewLine(){ Line line; line.direction = target - eye; line.passage = eye; return line; }
	Vector eyePosition(){ return eye; }
	void rotateRight( double angle ){
		Vector viewDirection;
		viewDirection.subtract(target, eye);
		viewDirection.rotate(zenith,angle);
		eye.subtract(target, viewDirection);
		this->lookAt(eye,target,up);
	}
	void rotateLeft( double angle ){ this->rotateRight( -angle ); }
	void Viewer::rotateDown( double angle ){
		Vector viewDirection;
		viewDirection.subtract(target, eye);
		viewDirection.rotate(this->rightDirection(),angle);
		eye.subtract(target, viewDirection);
		this->lookAt(eye,target,up);
	}
	void rotateUp( double angle ){ this->rotateDown( -angle ); }
	void twistRight( double angle ){ up.rotate(this->viewDirection(),angle); this->lookAt(eye,target,up); }
	void twistLeft( double angle ){ this->twistRight( -angle ); }
	void zoom( double scale ){
		Vector viewDirection;
		viewDirection.subtract(target, eye);
		viewDirection.multiply(scale);
		if(viewDirection.abs2()<1.0) viewDirection.normalize();
		eye.subtract(target, viewDirection);
		this->lookAt(eye,target,up);
	}
	void slideRight( double delta ){
		Vector rightDirection;
		rightDirection.exterior(this->viewDirection(), up);
		rightDirection.normalize();
		rightDirection.multiply(delta);
		eye.add(rightDirection);
		target.add(rightDirection);
		this->lookAt(eye,target,up);
	}
	void slideLeft( double delta ){ this->slideRight( -delta ); }
};

#endif //_VIEWER_H
